/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002 bazyl
   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   showing e-mail in scrolling window
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include "ecurses.h"
#include "mailreader.h"
#include "status.h"
#include "error.h"
#include "xmalloc.h"
#include "attach.h"
#include "cmd.h"
#include "wrapbox.h"
#include "gettext.h"
#include "folder.h"
#include "select.h"
#include "color.h"
#include "label.h"
#include "interface.h"
#include "ask.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/* This tells us if user is watching a header or a message. */
static int header_view = 0;

/* Mailreader window consists of select_t object and an optional label. */
static elabel_t *label       = NULL;
static select_t *mail_select = NULL;

/* Message body and header split into single lines. */
static rstring_t *header_lines = NULL;
static rstring_t *mail_lines   = NULL;

/* Message body as a string. */
static char *mail_body = NULL;

/* Colors used in mailreader window. */
static chtype text_color;
static chtype header_color;
static chtype signature_color;
static chtype quote_color;

/* This is the index of the first line of the signature. */
static unsigned signature_index = -1;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static rstring_t *
generate_header (mail_t *mail)
{
        rstring_t *result = rstring_create ();
        int        i;

        result->allocated_all = 1;
        
        if (mail->date_str)
                rstring_sprintf (result, "Date: %s", mail->date_str);
        if (mail->from)
                rstring_sprintf (result, "From: %s", mail->from->full);
        if (mail->subject)
                rstring_sprintf (result, "Subject: %s", mail->subject);
        if (mail->to && mail->to->count){
                if (mail->to->array[0] && mail->to->array[0]->full)
                        rstring_sprintf (result, "To: %s",
                                         mail->to->array[0]->full);
                else
                        rstring_sprintf (result, "To:");
                for (i = 1; i < mail->to->count; i++){
                        rstring_sprintf (result, "    %s",
                                         mail->to->array[i]->full);
                }
        }
        if (mail->cc && mail->cc->count){
                if (mail->cc->array[0] && mail->cc->array[0]->full)
                        rstring_sprintf (result, "Cc: %s",
                                         mail->cc->array[0]->full);
                else
                        rstring_sprintf (result, "Cc:");
                for (i = 1; i < mail->cc->count; i++){
                        rstring_sprintf (result, "    %s",
                                         mail->cc->array[i]->full);
                }
        }
        if (mail->bcc && mail->bcc->count){
                if (mail->bcc->array[0] && mail->bcc->array[0]->full)
                        rstring_sprintf (result, "Bcc: %s",
                                         mail->bcc->array[0]->full);
                else
                        rstring_sprintf (result, "Bcc:");
                for (i = 1; i < mail->bcc->count; i++){
                        rstring_sprintf (result, "    %s",
                                         mail->bcc->array[i]->full);
                }
        }
        if (mail->reply_to && mail->reply_to != mail->from)
                rstring_sprintf (result, "Reply-To: %s",
                                 mail->reply_to->full);
        if (mail->mua){
                rstring_sprintf (result, "MUA: %s", mail->mua);
        }
        rstring_sprintf (result, "%s", "");
        return result;
}



static int
set_mail_lines (int mime_index)
{
        int     width, nothing;
        str_t  *str;
        mail_t *mail = folder_mail_selected ();
        
        if (mail == NULL)
                return 1;
        
        if (mail_body){
                xfree (mail_body);
                mail_body = NULL;
        }

        str = wrapbox_mail_body (mail, mime_index);
        if (str == NULL && errno != 0){
                error_ (errno, _("impossible to display mail contents"));
                return 1;
        }
        mail_body = (str) ? str_finished (str) : NULL;

        if (mail_lines){
                rstring_delete (mail_lines);
                mail_lines = NULL;
        }

        if (header_lines){
                rstring_delete (header_lines);
                header_lines = NULL;
        }
  
        header_lines = generate_header (mail);

        getmaxyx (mail_select->win, nothing, width);
        
        if (mail_body){
                mail_lines = rstring_split_lines (mail_body, width);
        }
        return 0;
}



static int
set_header_lines (void)
{
        mail_t *mail = folder_mail_selected ();

        if (mail == NULL)
                return 1;
        
        if (mail_body){
                xfree (mail_body);
                mail_body = NULL;
        }
        
        if (wrapbox_mail_header (mail, &mail_body)){
                error_ (errno, _("impossible to display mail header"));
                return 1;
        }

        if (mail_lines){
                rstring_delete (mail_lines);
                mail_lines = NULL;
        }

        if (header_lines){
                rstring_delete (header_lines);
                header_lines = NULL;
        }
        
        if (mail_body){
                mail_lines = rstring_split_lines (mail_body, COLS);
        }
        return 0;
}



static int
quote_delta (int state, char c)
{
        if (state == 0 && isspace (c))
                return 0;

        if (c == '>')
                return 111;

        if (isalnum (c))
                return state + 1;

        return 5;
}


static int
is_quote (char *str)
{
        int state = 0;

        while (*str && state < 4){
                state = quote_delta (state, *str);
                str++;
        }

        return state == 111;
}



static int
sig_delta (int state, char c)
{
        if (c == '-')
                return state + 1;

        if (isspace (c))
                return state;

        return 3;
}


static int
is_signature (char *str, int index)
{
        int state = 0;
        
        if (index >= signature_index)
                return 1;

        while (*str && state < 3){
                state = sig_delta (state, *str);
                str++;
        }

        if (*str == '\0' && state == 2){
                signature_index = index;
                return 1;
        }

        return 0;
}



static void
change_color (WINDOW *window, char *str, int index)
{
        if (is_signature (str, index)){
                wattrset (window, signature_color);
        }
        else if (is_quote (str)){
                wattrset (window, quote_color);
        }
        else {
                wattrset (window, text_color);
        }
}


static void
draw_line (WINDOW *window, int maxlen, int index)
{
        int i;
        
        if (header_lines == NULL && mail_lines == NULL)
                return;

        if (header_lines && index < header_lines->count){
                wattrset (window, header_color);
                i = window_addnstr (window, header_lines->array[index],
                                    maxlen);
                while (i < maxlen){
                        window_addch (window, ' ');
                        i++;
                }
                return;
        }
        
        if (header_lines)
                index -= header_lines->count;
        
        if (mail_lines && index < mail_lines->count){
                change_color (window, mail_lines->array[index], index);
                i = window_addnstr (window, mail_lines->array[index],
                                    maxlen);
                while (i < maxlen){
                        window_addch (window, ' ');
                        i++;
                }
        }
        else {
                for (i = 0; i < maxlen; i++){
                        window_addch (window, ' ');
                }
        }
}


static int
reader_count (select_t *nothing)
{
        int ret = 1;
        
        if (header_lines)
                ret += header_lines->count;
        
        if (mail_lines)
                ret += mail_lines->count;

        return ret;
}


/* This file is generated by interface.pl script from interface.desc,
   and inc.in. */
static WINDOW *interface_init (void);
#include "mailreader.inc"

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
mailreader_init (void)
{
        WINDOW *window;

        window      = interface_init ();
        mail_select = select_open (window, 1, draw_line, reader_count);
        header_view = 0;

        window_set_functions (window, mailreader_refresh, mailreader_redraw,
                              mailreader_set_focus, mailreader_unset_focus);
}



void
mailreader_free_resources (void)
{
        if (mail_select)
                select_close (mail_select);
        mail_select = NULL;

        if (mail_lines)
                rstring_delete (mail_lines);
        mail_lines = NULL;

        if (header_lines)
                rstring_delete (header_lines);
        header_lines = NULL;

        if (mail_body)
                xfree (mail_body);
        mail_body = NULL;

        if (label)
                label_destroy (label);
        label = NULL;
}



void
mailreader_line_up (void)
{
        select_prev (mail_select);
}



void
mailreader_line_down (void)
{
        select_next (mail_select);
}



void
mailreader_page_up (void)
{
        select_prev_page (mail_select);
}



void
mailreader_page_down (void)
{
        select_next_page (mail_select);
}



void
mailreader_top (void)
{
        select_first (mail_select);
}



void
mailreader_bottom (void)
{
        select_last (mail_select);
}



void
mailreader_header_switch (void)
{
        if (! header_view){
                set_header_lines ();
                mailreader_redraw ();
                header_view = 1;
        }
        else {
                set_mail_lines (-1);
                mailreader_redraw ();
                header_view = 0;
        }
}



void
mailreader_show (void)
{
        select_first (mail_select);
        set_mail_lines (-1);
        header_view = 0;

        folder_after_read ();
        
        signature_index = -1;
        
        window_show (mail_select->win);
}


void
mailreader_close (void)
{
        signature_index = -1;
        
        if (mail_lines)
                rstring_delete (mail_lines);
        mail_lines = 0;
        
        window_hide (mail_select->win);

        interface_focus (1);
}



void
mailreader_redraw (void)
{
        mail_t *mail;
        
        mail = folder_mail_selected ();

        if (mail == NULL)
                return;
        
        select_redraw (mail_select);

        if (label)
                label_redraw (label);

        status_put_mailinfo (address_name (mail->from), mail->subject);
}



void
mailreader_refresh (void)
{
        select_show (mail_select);

        if (label)
                label_show (label);

        status_refresh ();
}


void
mailreader_set_focus (void)
{
        if (label){
                label_set_focus (label);
        }

        cmd_state_push (CMD_READ_MAIL);

        mailreader_redraw ();
}


void
mailreader_unset_focus (void)
{
        if (label){
                label_unset_focus (label);
                label_redraw (label);
        }

        cmd_state_pop ();
}


void
mailreader_change_mime (int index)
{
        mail_t *mail;

        mail = folder_mail_selected ();

        if (mail == NULL)
                return;
        
        if (index < 0 || index >= mail->rmime->count)
                return;

        set_mail_lines (index);
        mailreader_redraw ();
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE mailreader.c
 *
 ****************************************************************************/
